from typing import Optional

import mock
import torch
from torch import Tensor
from torch.nn import Sequential as Seq, Linear as Lin, ReLU
from torch_scatter import scatter_mean
from torch_geometric.nn import MetaLayer


def test_meta_layer():
    assert MetaLayer().__repr__() == ('MetaLayer(\n'
                                      '    edge_model=None,\n'
                                      '    node_model=None,\n'
                                      '    global_model=None\n'
                                      ')')

    edge_model = mock.MagicMock()
    node_model = mock.MagicMock()
    global_model = mock.MagicMock()

    for em in (edge_model, None):
        for nm in (node_model, None):
            for gm in (global_model, None):
                model = MetaLayer(em, nm, gm)
                out = model(mock.MagicMock(), edge_index=("row", "col"),
                            edge_attr="edge_attr", u="u",
                            batch=mock.MagicMock())

                assert isinstance(out, tuple) and len(out) == 3

                if em is not None:
                    em.assert_called_once()
                if nm is not None:
                    nm.assert_called_once()
                if gm is not None:
                    gm.assert_called_once()

                edge_model.reset_mock()
                node_model.reset_mock()
                global_model.reset_mock()


def test_meta_layer_example():
    class EdgeModel(torch.nn.Module):
        def __init__(self):
            super(EdgeModel, self).__init__()
            self.edge_mlp = Seq(Lin(2 * 10 + 5 + 20, 5), ReLU(), Lin(5, 5))

        def forward(self, src: Tensor, dest: Tensor,
                    edge_attr: Optional[Tensor], u: Optional[Tensor],
                    batch: Optional[Tensor]) -> Tensor:
            assert edge_attr is not None
            assert u is not None
            assert batch is not None
            out = torch.cat([src, dest, edge_attr, u[batch]], 1)
            return self.edge_mlp(out)

    class NodeModel(torch.nn.Module):
        def __init__(self):
            super(NodeModel, self).__init__()
            self.node_mlp_1 = Seq(Lin(15, 10), ReLU(), Lin(10, 10))
            self.node_mlp_2 = Seq(Lin(2 * 10 + 20, 10), ReLU(), Lin(10, 10))

        def forward(self, x: Tensor, edge_index: Tensor,
                    edge_attr: Optional[Tensor], u: Optional[Tensor],
                    batch: Optional[Tensor]) -> Tensor:
            assert edge_attr is not None
            assert u is not None
            assert batch is not None
            row = edge_index[0]
            col = edge_index[1]
            out = torch.cat([x[row], edge_attr], dim=1)
            out = self.node_mlp_1(out)
            out = scatter_mean(out, col, dim=0, dim_size=x.size(0))
            out = torch.cat([x, out, u[batch]], dim=1)
            return self.node_mlp_2(out)

    class GlobalModel(torch.nn.Module):
        def __init__(self):
            super(GlobalModel, self).__init__()
            self.global_mlp = Seq(Lin(20 + 10, 20), ReLU(), Lin(20, 20))

        def forward(self, x: Tensor, edge_index: Tensor,
                    edge_attr: Optional[Tensor], u: Optional[Tensor],
                    batch: Optional[Tensor]) -> Tensor:
            assert u is not None
            assert batch is not None
            out = torch.cat([u, scatter_mean(x, batch, dim=0)], dim=1)
            return self.global_mlp(out)

    op = MetaLayer(EdgeModel(), NodeModel(), GlobalModel())

    x = torch.randn(20, 10)
    edge_attr = torch.randn(40, 5)
    u = torch.randn(2, 20)
    batch = torch.tensor([0] * 10 + [1] * 10)
    edge_index = torch.randint(0, high=10, size=(2, 20), dtype=torch.long)
    edge_index = torch.cat([edge_index, 10 + edge_index], dim=1)

    x, edge_attr, u, op(x, edge_index, edge_attr, u, batch)
    assert x.size() == (20, 10)
    assert edge_attr.size() == (40, 5)
    assert u.size() == (2, 20)

    torch.jit.script(op)
